
#
#
# TODO: Better keyboard handling
# TODO: Add more rectangle decoding options
# TODO: Add more security handshake options (all of the RFC is already implemented, so only custom are to be added)
# TODO: Add mouse scroll functionality (QT5 client needs to be modified for that)

import io
import asyncio
import traceback
from struct import pack, unpack
from typing import cast

from aardwolf import logger
from aardwolf.transport.tcpstream import TCPStream
from unicrypto.symmetric import DES, MODE_ECB

from aardwolf.commons.target import RDPTarget
from aardwolf.commons.queuedata import RDPDATATYPE, RDP_KEYBOARD_SCANCODE, RDP_KEYBOARD_UNICODE, \
	RDP_MOUSE, RDP_VIDEO, RDP_CLIPBOARD_READY, RDP_CLIPBOARD_DATA_TXT, RDP_CLIPBOARD_NEW_DATA_AVAILABLE, \
	RDP_BEEP
from aardwolf.extensions.RDPECLIP.protocol.formatlist import CLIPBRD_FORMAT
from aardwolf.protocol.vnc.keyboard import *
from aardwolf.keyboard import VK_MODIFIERS
from aardwolf.keyboard.layoutmanager import KeyboardLayoutManager
from aardwolf.commons.queuedata.constants import MOUSEBUTTON, VIDEO_FORMAT
from aardwolf.commons.iosettings import RDPIOSettings
from asyauth.common.credentials import UniCredential
from asyauth.common.constants import asyauthSecret
from asysocks.unicomm.client import UniClient
from asysocks.unicomm.common.connection import UniConnection
from asysocks.unicomm.common.packetizers import StreamPacketizer

from PIL import Image
try:
	from PIL.ImageQt import ImageQt
except ImportError:
	logger.debug('No Qt installed! Converting to qt will not work')

import librlers

# https://datatracker.ietf.org/doc/html/rfc6143

RAW_ENCODING = 0
COPY_RECTANGLE_ENCODING = 1
RRE_ENCODING = 2
CORRE_ENCODING = 4
HEXTILE_ENCODING = 5
ZLIB_ENCODING = 6
TIGHT_ENCODING = 7
ZLIBHEX_ENCODING = 8
TRLE_ENCODING = 15
ZRLE_ENCODING = 16

class VNCConnection:
	def __init__(self, target:RDPTarget, credentials:UniCredential, iosettings:RDPIOSettings):
		self.target = target
		self.credentials = credentials
		self.authapi = None
		self.iosettings = iosettings
		self.__connection:UniConnection = None
		self.shared_flag = False
		self.client_version = '003.008'
		self.server_version = None
		self.server_name = None
		self.disconnected_evt = asyncio.Event() #this will be set if we disconnect for whatever reason
		self.server_supp_security_types = []
		self.__selected_security_type = 1 if self.credentials.stype == asyauthSecret.NONE else 2 #currently we only support these 2
		self.__refresh_screen_task = None
		self.__reader_loop_task = None
		self.__external_reader_task = None
		self.__use_pyperclip = False

		# these are the main queues with which you can communicate with the server
		# ext_out_queue: yields video data
		# ext_in_queue: expects keyboard/mouse data
		self.ext_out_queue = asyncio.Queue()
		self.ext_in_queue = asyncio.Queue()

		self.__reader = None
		self.__writer = None


		self.bpp = None
		self.bypp = None
		self.depth = None
		self.bigendian = None
		self.truecolor = None 
		self.redmax = None 
		self.greenmax = None
		self.bluemax = None
		self.redshift = None 
		self.greenshift = None
		self.blueshift = None
		self.width = None
		self.height = None
		self.__desktop_buffer = None
		self.desktop_buffer_has_data = False

		self.__vk_to_vnckey = {
			'VK_BACK' : KEY_BackSpace,
			'VK_ESCAPE' : KEY_Escape,
			'VK_TAB' : KEY_Tab,
			'VK_RETURN' : KEY_Return,
			'VK_INSERT' : KEY_Insert,
			'VK_DELETE' : KEY_Delete,
			'VK_HOME' : KEY_Home,
			'VK_END' : KEY_End,
			'VK_PRIOR' : KEY_PageUp,
			'VK_NEXT' : KEY_PageDown,
			'VK_LEFT' : KEY_Left,
			'VK_UP' : KEY_Up,
			'VK_RIGHT' : KEY_Right,
			'VK_DOWN' : KEY_Down,
			'VK_HOME' : KEY_Home,
			'VK_F1' : KEY_F1,
			'VK_F2' : KEY_F2,
			'VK_F3' : KEY_F3,
			'VK_F4' : KEY_F4,
			'VK_F5' : KEY_F5,
			'VK_F6' : KEY_F6,
			'VK_F7' : KEY_F7,
			'VK_F8' : KEY_F8,
			'VK_F9' : KEY_F9,
			'VK_F10' : KEY_F10,
			'VK_F11' : KEY_F11,
			'VK_F12' : KEY_F12,
			#'VK_F13' : KEY_F13,
			#'VK_F14' : KEY_F14,
			#'VK_F15' : KEY_F15,
			#'VK_F16' : KEY_F16,
			#'VK_F17' : KEY_F17,
			#'VK_F18' : KEY_F18,
			#'VK_F19' : KEY_F19,
			#'VK_F20' : KEY_F20,
			'VK_LSHIFT' : KEY_ShiftLeft,
			'VK_RSHIFT' : KEY_ShiftRight,
			'VK_LCONTROL' : KEY_ControlLeft,
			'VK_RCONTROL' : KEY_ControlRight,
			'VK_LWIN' : KEY_Super_L,
			'VK_RWIN' : KEY_Super_R,
			'VK_LMENU' : KEY_AltLeft,
			'VK_RMENU' : KEY_AltRight,
			'VK_SCROLL' : KEY_Scroll_Lock,
			#'????' : KEY_Sys_Req,
			'VK_NUMLOCK' : KEY_Num_Lock,
			'VK_CAPITAL' : KEY_Caps_Lock,
			'VK_PAUSE' : KEY_Pause,
			#'????' : KEY_Super_L,
			#'????' : KEY_Super_R,
			#'????' : KEY_Hyper_L,
			#'????' : KEY_Hyper_R,
			#'????' : KEY_KP_0,
			#'????' : KEY_KP_1,
			#'????' : KEY_KP_2,
			#'????' : KEY_KP_3,
			#'????' : KEY_KP_4,
			#'????' : KEY_KP_5,
			#'????' : KEY_KP_6,
			#'????' : KEY_KP_7,
			#'????' : KEY_KP_8,
			#'????' : KEY_KP_9,
			#'????': KEY_KP_Enter,
			'VK_MULTIPLY': KEY_KP_Multiply,
			'VK_ADD': KEY_KP_Add,
			'VK_SUBTRACT': KEY_KP_Subtract,
			'VK_DECIMAL' : KEY_KP_Decimal,
			'VK_DIVIDE' : KEY_KP_Divide,
			'VK_SNAPSHOT' : KEY_Print,
			'VK_DBE_NOCODEINPUT' : KEY_KP_Divide,
		}

		self.__keyboard_layout = KeyboardLayoutManager().get_layout_by_shortname(self.iosettings.client_keyboard)
	
	async def terminate(self):
		try:
			if self.__writer is not None:
				await self.__writer.close()
			if self.__refresh_screen_task is not None:
				self.__refresh_screen_task.cancel()
			if self.__reader_loop_task is not None:
				self.__reader_loop_task.cancel()
			if self.__external_reader_task is not None:
				self.__external_reader_task.cancel()
			
			return True, None
		except Exception as e:
			traceback.print_exc()
			return None, e
		finally:
			self.disconnected_evt.set()
	
	async def __aenter__(self):
		return self
		
	async def __aexit__(self, exc_type, exc, traceback):
		await asyncio.wait_for(self.terminate(), timeout = 5)
	
	def get_extra_info(self):
		# to have the same interface as RDP
		return None
	
	async def connect(self):
		"""
		Performs the entire connection sequence 
		"""
		try:
			packetizer = StreamPacketizer()
			client = UniClient(self.target, packetizer)
			self.__connection = await client.connect()
			asyncio.create_task(self.__connection.stream())
			self.__reader = self.__connection.packetizer
			self.__writer = self.__connection

			logger.debug('Performing banner exchange')
			_, err = await self.__banner_exchange()
			if err is not None:
				raise err
			logger.debug('Banner exchange OK')

			logger.debug('Performing security handshake')
			_, err = await self.__security_handshake()
			if err is not None:
				raise err
			logger.debug('Security handshake OK')

			logger.debug('Authenticating')
			_, err = await self.__authenticate()
			if err is not None:
				raise err
			logger.debug('Authentication OK')

			logger.debug('Setting up clipboard')
			_, err = await self.__setup_clipboard()
			if err is not None:
				raise err
			
			logger.debug('Client init')
			_, err = await self.__client_init()
			if err is not None:
				raise err
			logger.debug('Client init OK')

			logger.debug('Starting internal comm channels')
			self.__reader_loop_task = asyncio.create_task(self.__reader_loop())
			self.__external_reader_task = asyncio.create_task(self.__external_reader())
			logger.debug('Sending cliboard ready signal') #to emulate RDP clipboard functionality
			msg = RDP_CLIPBOARD_READY()
			await self.ext_out_queue.put(msg)
			
			return True, None
		except Exception as e:
			await self.terminate()
			return None, e

	async def __setup_clipboard(self):
		if self.iosettings.clipboard_use_pyperclip is True:
			try:
				import pyperclip
			except ImportError:
				logger.debug('Could not import pyperclip! Copy-paste will not work!')
				self.__use_pyperclip = False
			else:
				self.__use_pyperclip = True
				if not pyperclip.is_available():
					logger.debug("pyperclip - Copy functionality available!")
		return True, None
	
	async def __banner_exchange(self):
		try:
			banner = await self.__reader.readuntil(b'\n')
			self.server_version = banner[4:-1].decode()
			logger.debug('Server version: %s' % self.server_version)
			version_reply = b'RFB %s\n' % self.client_version.encode()
			logger.debug('Version reply: %s' % version_reply)
			await self.__writer.write(version_reply)
			return True, None
		except Exception as e:
			return None, e

	async def __security_handshake(self):
		try:
			no_sec_types = await self.__reader.readexactly(1)
			no_sec_types = ord(no_sec_types)
			if no_sec_types == 0:
				logger.debug('Server sent empty support security types, connection will terminate soon')
				err_string_size = await self.__reader.readexactly(1)
				err_string_size = ord(err_string_size)
				err_string = await self.__reader.readexactly(err_string_size)
				err_string = err_string.decode()
				raise Exception(err_string)
			sec_types = await self.__reader.readexactly(no_sec_types)
			for sectype in sec_types:
				self.server_supp_security_types.append(sectype)
			
			if self.__selected_security_type not in self.server_supp_security_types:
				raise Exception('Clound\'t find common authentication type. Client supports: %s Server supports: %s' % (self.__selected_security_type, ','.join([str(x) for x in self.server_supp_security_types])))


			return True, None
		except Exception as e:
			return None, e

	async def __authenticate(self):
		try:
			if self.__selected_security_type == 0:
				logger.debug('Invalid authentication type!!!')
				raise Exception('Invalid authentication type')
			
			elif self.__selected_security_type == 1:
				# nothing to do here
				logger.debug('Selecting NULL auth type')

			
			elif self.__selected_security_type == 2:
				logger.debug('Selecting default VNC auth type')
				await self.__writer.write(bytes([self.__selected_security_type]))
				challenge = await self.__reader.readexactly(16)

				# no username used here, but we support RDP as well
				password = self.credentials.secret
				if self.credentials.secret is None:
					password = self.credentials.username
				
				password = password.ljust(8, '\x00').encode('ascii')
				password = password[:8]
				# converting password to key
				newkey = b''
				for ki in range(len(password)):
					bsrc = password[ki]
					btgt = 0
					for i in range(8):
						if bsrc & (1 << i):
							btgt = btgt | (1 << 7 - i)
					newkey+= bytes([btgt])
				ctx = DES(newkey, mode = MODE_ECB, IV = None)
				response = ctx.encrypt(challenge)
				await self.__writer.write(response)

			else:
				raise Exception('Unsupported security type selected!')
			
			auth_result = await self.__reader.readexactly(4)
			auth_result = int.from_bytes(auth_result, byteorder = 'big', signed=False)
			if auth_result != 0:
				logger.debug('Auth Failed! Auth result: %s' % auth_result)
				err_string_size = await self.__reader.readexactly(4)
				err_string_size = int.from_bytes(err_string_size, byteorder = 'big', signed=False)
				err_string = await self.__reader.readexactly(err_string_size)
				err_string = err_string.decode()
				raise Exception(err_string)
			logger.debug('Auth OK!')
			return True, None
		except Exception as e:
			return None, e
	
	async def __client_init(self):
		try:
			# sending client init
			await self.__writer.write(bytes([int(self.shared_flag)]))
			# reading server_init
			framebuffer_width = await self.__reader.readexactly(2)
			self.width = int.from_bytes(framebuffer_width, byteorder = 'big', signed = False)
			framebuffer_height = await self.__reader.readexactly(2)
			self.height = int.from_bytes(framebuffer_height, byteorder = 'big', signed = False)
			pixel_format_raw = await self.__reader.readexactly(16)
			self.bpp, self.depth, self.bigendian, self.truecolor, \
			self.redmax, self.greenmax, self.bluemax, self.redshift, \
			self.greenshift, self.blueshift = unpack("!BBBBHHHBBBxxx", pixel_format_raw)
			self.bypp = self.bpp // 8  # calc bytes per pixel

			name_string_size = await self.__reader.readexactly(4)
			name_string_size = int.from_bytes(name_string_size, byteorder = 'big', signed=False)
			name_string = await self.__reader.readexactly(name_string_size)
			self.server_name = name_string.decode()


			self.__desktop_buffer = Image.new(mode="RGBA", size=(self.width, self.height))
			_, err = await self.set_encodings(self.iosettings.vnc_encodings)
			if err is not None:
				raise err
			
			# don't change this!!!!
			await self.set_pixel_format()
			await self.framebuffer_update_request()
			
			self.__refresh_screen_task = asyncio.create_task(self.__refresh_screen_loop())
			return True, None
		except Exception as e:
			return None, e
	
	async def send_key_virtualkey(self, vk:str, is_pressed:bool, is_extended:bool, scancode_hint:int = None, modifiers = VK_MODIFIERS(0)):
		try:
			if vk is None:
				return await self.send_key_scancode(scancode_hint, is_pressed, is_extended, modifiers=modifiers)
			#print('Got VK: %s' % vk)
			if vk in self.__vk_to_vnckey:
				keycode = self.__vk_to_vnckey[vk]
				return await self.send_key_char(keycode, is_pressed)
			else:
				return await self.send_key_scancode(scancode_hint, is_pressed, is_extended, modifiers=modifiers)
		except Exception as e:
			traceback.print_exc()
			return None, e

	async def send_key_scancode(self, scancode, is_pressed, is_extended, modifiers = VK_MODIFIERS(0)):
		try:
			keycode = None
			if modifiers == VK_MODIFIERS(0):
				vk = self.__keyboard_layout.scancode_to_vk(scancode)
				if vk is not None and vk in self.__vk_to_vnckey:
					keycode = self.__vk_to_vnckey[vk]
			if keycode is None:
				keycode = self.__keyboard_layout.scancode_to_char(scancode, modifiers)
				if keycode is None:
					raise Exception('Failed to resolv key! SC: %s MOD: %s' % (scancode, modifiers))

				elif keycode is not None and len(keycode) == 1:
					keycode = ord(keycode)
				elif keycode is not None and len(keycode) > 1:
					raise Exception('LARGE! Keycode %s resolved to: %s' % (scancode , repr(keycode)))
				else:
					raise Exception('This key is too special! Can\'t resolve it! SC: %s' % scancode)

			return await self.send_key_char(keycode, is_pressed)
		except Exception as e:
			traceback.print_exc()
			return None, e

	async def send_key_char(self, char, is_pressed):
		try:
			msg = pack("!BBxxI", 4, int(is_pressed), char)
			await self.__writer.write(msg)
			return True, None
		except Exception as e:
			traceback.print_exc()
			return None, e

	async def send_mouse(self, button:MOUSEBUTTON, xPos:int, yPos:int, is_pressed:bool):
		try:
			if xPos < 0 or yPos < 0:
				return True, None

			buttoncode = 0
			if button == MOUSEBUTTON.MOUSEBUTTON_LEFT:
				buttoncode = 1
			elif button == MOUSEBUTTON.MOUSEBUTTON_MIDDLE:
				buttoncode = 2
			elif button == MOUSEBUTTON.MOUSEBUTTON_RIGHT:
				buttoncode = 3
			
			buttonmask = 0
			if is_pressed is True:
				if buttoncode == 1: buttonmask &= ~1
				if buttoncode == 2: buttonmask &= ~2
				if buttoncode == 3: buttonmask &= ~4
				if buttoncode == 4: buttonmask &= ~8
				if buttoncode == 5: buttonmask &= ~16
			else:
				if buttoncode == 1: buttonmask |= 1
				if buttoncode == 2: buttonmask |= 2
				if buttoncode == 3: buttonmask |= 4
				if buttoncode == 4: buttonmask |= 8
				if buttoncode == 5: buttonmask |= 16
			
			msg = pack("!BBHH", 5, buttonmask, xPos, yPos)
			await self.__writer.write(msg)
			return True, None
		except Exception as e:
			traceback.print_exc()
			return None, e

	def get_desktop_buffer(self, encoding:VIDEO_FORMAT = VIDEO_FORMAT.PIL):
		"""Makes a copy of the current desktop buffer, converts it and returns the object"""
		try:
			image = self.__desktop_buffer.copy()
			if encoding == VIDEO_FORMAT.PIL:
				return image
			elif encoding == VIDEO_FORMAT.RAW:
				return image.tobytes()
			elif encoding == VIDEO_FORMAT.QT5:
				return ImageQt(image)
			elif encoding == VIDEO_FORMAT.PNG:
				img_byte_arr = io.BytesIO()
				image.save(img_byte_arr, format='PNG')
				return img_byte_arr.getvalue()
			else:
				raise ValueError('Output format of "%s" is not supported!' % encoding)
		except Exception as e:
			traceback.print_exc()
			return None, e

	async def __external_reader(self):
		# This coroutine handles keyboard/mouse/clipboard etc input from the user
		# It wraps the data in it's appropriate format then dispatches it to the server
		try:
			while not self.disconnected_evt.is_set():
				indata = await self.ext_in_queue.get()
				if indata is None:
					logger.debug('External queue got None, upper layer is exiting!')
					await self.terminate()
					return
				if indata.type == RDPDATATYPE.KEYSCAN:
					indata = cast(RDP_KEYBOARD_SCANCODE, indata)
					if indata.vk_code is not None:
						await self.send_key_virtualkey(indata.vk_code, indata.is_pressed, indata.is_extended, scancode_hint=indata.keyCode, modifiers=indata.modifiers)
					else:
						await self.send_key_scancode(indata.keyCode, indata.is_pressed, indata.is_extended, modifiers=indata.modifiers)
					
					
					
					
					#modifiers = indata.modifiers
					#
					### emulating keys...
					#keycode = None
					#try:
					#	if indata.keyCode is None and indata.vk_code is None:
					#		print('No keycode found! ')
					#		continue
					#
					#	if indata.vk_code is not None:
					#		vk_code = indata.vk_code
					#	else:
					#		vk_code = self.__keyboard_layout.scancode_to_vk(indata.keyCode)
					#	print('Got VK: %s' % vk_code)
					#	if vk_code is None:
					#		print('Could not map SC to VK! SC: %s' % indata.keyCode)
					#	if vk_code is not None and vk_code in self.__vk_to_vnckey:
					#		keycode = self.__vk_to_vnckey[vk_code]
					#		print('AAAAAAAA %s' % hex(keycode))
					#
					#	if keycode is None:
					#		keycode = self.__keyboard_layout.scancode_to_char(indata.keyCode, modifiers)
					#		print(keycode)
					#		if keycode is None:
					#			print('Failed to resolv key! SC: %s VK: %s' % (indata.keyCode, vk_code))
					#			continue
					#		elif keycode is not None and len(keycode) == 1:
					#			keycode = ord(keycode)
					#			print('Keycode %s resolved to: %s' % (indata.keyCode , repr(keycode)))
					#		elif keycode is not None and len(keycode) > 1:
					#			print('LARGE! Keycode %s resolved to: %s' % (indata.keyCode , repr(keycode)))
					#			continue
					#		else:
					#			print('This key is too special! Can\'t resolve it! SC: %s VK: %s' % (indata.keyCode, vk_code))
					#			continue
					#	
					#except Exception as e:
					#	traceback.print_exc()
					#	continue
					#
					#if indata.keyCode is not None:
					#	print('Original kk  : %s [%s]' % (indata.keyCode, hex(indata.keyCode)))
					#print('Final keycode: %s' % hex(keycode))
					#print('Is pressed   : %s' % indata.is_pressed)
					#if keycode is not None:
					#	msg = pack("!BBxxI", 4, int(indata.is_pressed), keycode)
					#	self.__writer.write(msg)
				
				elif indata.type == RDPDATATYPE.KEYUNICODE:
					indata = cast(RDP_KEYBOARD_UNICODE, indata)
					keycode = indata.char
					if not isinstance(keycode, int):
						if len(keycode) == 1:
							keycode = ord(keycode)
						else:
							# I hope you know what you're doing here...
							keycode = int.from_bytes(bytes.fromhex(keycode), byteorder = 'big', signed = False)
					await self.send_key_char(keycode, indata.is_pressed)
			
				elif indata.type == RDPDATATYPE.MOUSE:
					#PointerEvent
					indata = cast(RDP_MOUSE, indata)
					if indata.xPos < 0 or indata.yPos < 0:
						continue
					
					await self.send_mouse(indata.button, indata.xPos, indata.yPos, indata.is_pressed)

				elif indata.type == RDPDATATYPE.CLIPBOARD_DATA_TXT:
					try:
						indata = cast(RDP_CLIPBOARD_DATA_TXT, indata)
						txtdata = indata.data.encode('latin-1')
						msg = pack("!BxxxI", 6, len(txtdata))
						msg += txtdata
						await self.__writer.write(msg)
					except Exception as e:
						traceback.print_exc()
						continue

		except asyncio.CancelledError:
			return None, None

		except Exception as e:
			traceback.print_exc()
			await self.terminate()
			return None, e

	async def __refresh_screen_loop(self):
		# Periodically sends a framebufferupdate request to the server
		# Potentially not the best way to do it

		while not self.disconnected_evt.is_set():
			try:
				await self.framebuffer_update_request(incremental=1)
				await asyncio.sleep(1/self.iosettings.vnc_fps)
			except asyncio.CancelledError:
				return
			except Exception as e:
				break

	async def __send_rect(self, x, y, width, height, image:Image):
		try:
			#updating desktop buffer to have a way to copy rectangles later
			self.desktop_buffer_has_data = True
			if self.width == width and self.height == height:
				self.__desktop_buffer = image
			else:
				self.__desktop_buffer.paste(image, [x, y, x+width, y+height])
			
			if self.iosettings.video_out_format == VIDEO_FORMAT.RAW:
				image = image.tobytes()

			elif self.iosettings.video_out_format == VIDEO_FORMAT.QT5:
				image = ImageQt(image)
			
			elif self.iosettings.video_out_format == VIDEO_FORMAT.PNG:
				img_byte_arr = io.BytesIO()
				image.save(img_byte_arr, format='PNG')
				image = img_byte_arr.getvalue()
			else:
				raise ValueError('Output format of "%s" is not supported!' % self.iosettings.video_out_format)
									
			rect = RDP_VIDEO()
			rect.x = x
			rect.y = y
			rect.width = width
			rect.height = height
			rect.bitsPerPixel = self.bpp
			rect.is_compressed = False
			rect.data = image
			await self.ext_out_queue.put(rect)

		except Exception as e:
			await self.terminate()
			return None, e

	async def __reader_loop(self):
		try:
			while True:
				msgtype = await self.__reader.readexactly(1)
				msgtype = ord(msgtype)
				if msgtype == 0:
					# Framebufferupdate message from the server
					# 
					num_rect = await self.__reader.readexactly(3)
					num_rect = int.from_bytes(num_rect[1:], byteorder = 'big', signed = False)
					for _ in range(num_rect):
						rect_hdr = await self.__reader.readexactly(12)
						(x, y, width, height, encoding) = unpack("!HHHHI", rect_hdr)
						if encoding == RAW_ENCODING:
							try:
								data = await self.__reader.readexactly(width*height*self.bypp)
								data = librlers.mask_rgbx(data)
								image = Image.frombytes('RGBA', [width, height], bytes(data))
								await self.__send_rect(x,y,width, height, image)
							except Exception as e:
								traceback.print_exc()
								return

						elif encoding == COPY_RECTANGLE_ENCODING:
							try:
								data = await self.__reader.readexactly(4)
								(srcx, srcy) = unpack("!HH", data)								
								newrect = self.__desktop_buffer.crop([srcx, srcy, srcx+width, srcy+height])
								await self.__send_rect(x,y,width,height, newrect)
								continue
							except Exception as e:
								traceback.print_exc()
								return


						elif encoding == RRE_ENCODING:
							try:
								raw_data = b''
								sub_rect_num = await self.__reader.readexactly(4)
								raw_data += sub_rect_num
								sub_rect_num = int.from_bytes(sub_rect_num, byteorder = 'big', signed = False)
								backgroud_pixel = await self.__reader.readexactly(self.bypp)
								raw_data += backgroud_pixel
								
								if sub_rect_num > 0:
									subrect_size = (self.bypp + 8)*sub_rect_num
									data = await self.__reader.readexactly(subrect_size)
									raw_data += data
								
								raw_out = librlers.decode_rre(raw_data, width, height, self.bypp)
								raw_out = bytes(raw_out)
								
								rect = Image.frombytes('RGBA', [width, height], raw_out)
								await self.__send_rect(x,y, width, height, rect)

								#### This is the pure-python version of RRE, kept here as a reminder
								#### It's a bit slow compared to the C implementation
								##
								## sub_rect_num = await self.__reader.readexactly(4)
								## sub_rect_num = int.from_bytes(sub_rect_num, byteorder = 'big', signed = False)
								## backgroud_pixel = await self.__reader.readexactly(self.bypp)
								## r,g,b,a = unpack("BBBB", backgroud_pixel)
								## a = 255
								## rect = Image.new('RGBA', [width, height])
								## rect.paste((r,g,b,a), [0,0, width, height])
								## 
								## format = "!%dsHHHH" % self.bypp
								## for _ in range(sub_rect_num):
								## 	data = await self.__reader.readexactly(self.bypp + 8)
								## 	(color, subx, suby, subwidth, subheight) = unpack(format, data)
								## 	r,g,b,a = unpack("BBBB", color)
								## 	a = 255
								## 	rect.paste((r,g,b,a), [subx,suby, subx+subwidth, suby+subheight])
								## await self.__send_rect(x,y, width, height, rect)
							except Exception as e:
								raise e
						
						
						elif encoding == TRLE_ENCODING:
							raise NotImplementedError()

						else:
							raise Exception('Unknown encoding %s' % encoding)

				elif msgtype == 1:
					# Colormap messages are only used in certain encoding formats, 
					# none of which are currently implemented
					# but the decoding process works
					hdr = await self.__reader.readexactly(5)
					(_, firstcolor, numcolor) = unpack("!BHH", hdr)
					for _ in range(numcolor):
						color_raw = await self.__reader.readexactly(6)
						(red, green, blue) = unpack("!HHH", color_raw)
				
				elif msgtype == 2:
					# Server sent a Bell command, should cause a beep
					msg = RDP_BEEP()
					await self.ext_out_queue.put(msg)
				
				elif msgtype == 3:
					# Server side has updated the clipboard
					
					# Signaling clipboard data (to simulate RDP functionality)
					msg = RDP_CLIPBOARD_NEW_DATA_AVAILABLE()
					await self.ext_out_queue.put(msg)

					# processing clipboard data
					hdr = await self.__reader.readexactly(7)
					(_,_,_, cliplen ) = unpack("!BBBI", hdr)
					cliptext = await self.__reader.readexactly(cliplen)
					cliptext = cliptext.decode('latin-1') #latin-1 is per RFC
					logger.debug('Got clipboard test: %s' % repr(cliptext))
					if self.__use_pyperclip is True:
						import pyperclip
						pyperclip.copy(cliptext)
					
					msg = RDP_CLIPBOARD_DATA_TXT()
					msg.data = cliptext
					msg.datatype = CLIPBRD_FORMAT.CF_UNICODETEXT
					await self.ext_out_queue.put(msg)
				
				else:
					raise Exception('Unexpected message tpye %s' % msgtype)
					
			return True, None
		except Exception as e:
			await self.terminate()
			return None, e

	async def set_encodings(self, list_of_encodings = [2, 1, 0]):
		"""Sends a setencoding message to the server, indicating the types of image encodings we support"""
		try:
			if list_of_encodings is None:
				list_of_encodings = [2, 1, 0]
			enc_encodings = b''
			for encoding in list_of_encodings:
				enc_encodings += encoding.to_bytes(4, byteorder = 'big', signed = True)
			sendbuff = pack("!BxH", 2, len(list_of_encodings)) + enc_encodings
			await self.__writer.write(sendbuff)
			return True, None
		except Exception as e:
			return None, e
		

	async def set_pixel_format(self, bpp=32, depth=24, bigendian=0, truecolor=1, redmax=255, greenmax=255, bluemax=255, redshift=0, greenshift=8, blueshift=16):
		"""Sends a setpixelformat message to the server, letting the server know our preferred image data encodings"""

		pixformat = pack("!BBBBHHHBBBxxx", bpp, depth, bigendian, truecolor, redmax, greenmax, bluemax, redshift, greenshift, blueshift)
		await self.__writer.write(b'\x00\x00\x00\x00' + pixformat)
		# rember these settings
		self.bpp, self.depth, self.bigendian, self.truecolor = bpp, depth, bigendian, truecolor
		self.redmax, self.greenmax, self.bluemax = redmax, greenmax, bluemax
		self.redshift, self.greenshift, self.blueshift = redshift, greenshift, blueshift
		self.bypp = self.bpp // 8  # calc bytes per pixel
		# ~ print self.bypp

	async def framebuffer_update_request(self, x=0, y=0, width=None, height=None, incremental=0):
		"""Sends a framebufferupdaterequest message to the server, indicating the area to be updated"""
		if width is None:
			width = self.width - x

		if height is None:
			height = self.height - y

		await self.__writer.write(pack("!BBHHHH", 3, incremental, x, y, width, height))

async def amain():
	try:
		logger.setLevel(1)
		from aardwolf.commons.factory import RDPConnectionFactory
		from aardwolf.commons.iosettings import RDPIOSettings

		iosettings = RDPIOSettings()
		iosettings.video_out_format = VIDEO_FORMAT.RAW
		
		url = 'vnc+plain-password://alma:alma@10.10.10.102:5900'
		url = RDPConnectionFactory.from_url(url)
		connection = url.get_connection(iosettings)

		_, err = await connection.connect()
		if err is not None:
			raise err

		while True:
			await asyncio.sleep(10)

		return True, None
	except Exception as e:
		traceback.print_exc()
		return None, e

def main():
	asyncio.run(amain())

if __name__ == '__main__':
	main()



import io
import enum

from aardwolf.extensions.RDPECLIP.protocol.formatlist import CLIPBRD_FORMAT

class CLIPRDR_FORMAT_DATA_REQUEST:
	def __init__(self):
		self.requestedFormatId:CLIPBRD_FORMAT = None

	def to_bytes(self):
		if isinstance(self.requestedFormatId, CLIPBRD_FORMAT):
			t = self.requestedFormatId.value.to_bytes(4, byteorder='little', signed=False)
		else:
			t = self.requestedFormatId.to_bytes(4, byteorder='little', signed=False)
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return CLIPRDR_FORMAT_DATA_REQUEST.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = CLIPRDR_FORMAT_DATA_REQUEST()
		t = buff.read(4)
		try:
			msg.requestedFormatId = CLIPBRD_FORMAT(int.from_bytes(t, byteorder='little', signed=False))
		except:
			msg.requestedFormatId = int.from_bytes(t, byteorder='little', signed=False)
		return msg

	def __repr__(self):
		t = '==== CLIPRDR_FORMAT_DATA_REQUEST ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t


import io
import enum

class TS_WINDOW_LEVEL(enum.Enum):
	NOT_SUPPORTED = 0x00000000 #The client or server is not capable of supporting Windowing Alternate Secondary Drawing Orders.
	SUPPORTED = 0x00000001 #The client or server is capable of supporting Windowing Alternate Secondary Drawing Orders.
	SUPPORTED_EX = 0x00000002 #The client or server is capable of supporting Windowing Alternate Secondary Drawing Orders 

class TS_WINDOW_LIST_CAP_SET:
	def __init__(self):
		self.WndSupportLevel:TS_WINDOW_LEVEL = TS_WINDOW_LEVEL.NOT_SUPPORTED
		self.NumIconCaches:int = None
		self.NumIconCacheEntries:int = None

	def to_bytes(self):
		t = self.WndSupportLevel.value.to_bytes(4, byteorder='little', signed = False)
		t += self.NumIconCaches.to_bytes(1, byteorder='little', signed = False)
		t += self.NumIconCacheEntries.to_bytes(2, byteorder='little', signed = False)
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_WINDOW_LIST_CAP_SET.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_WINDOW_LIST_CAP_SET()
		msg.WndSupportLevel = TS_WINDOW_LEVEL(int.from_bytes(buff.read(4), byteorder='little', signed = False))
		msg.NumIconCaches = int.from_bytes(buff.read(1), byteorder='little', signed = False)
		msg.NumIconCacheEntries = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		return msg

	def __repr__(self):
		t = '==== TS_WINDOW_LIST_CAP_SET ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t


from aardwolf.extensions.RDPECLIP.channel import RDPECLIPChannel
from aardwolf.protocol.x224.constants import SUPP_PROTOCOLS, NEG_FLAGS
from aardwolf.commons.queuedata.constants import VIDEO_FORMAT
from aardwolf.protocol.T125.extendedinfopacket import PERF
from aardwolf.extensions.RDPEDYC.channel import RDPEDYCChannel

from aardwolf.extensions.RDPEDYC.vchannels.echo import VchannelECHO
#from aardwolf.extensions.RDPEDYC.vchannels.test import VchannelTEST

class RDPIOSettings:
	def __init__(self):
		# RDP only settings
		# Which channels should be enabled
		self.channels = [RDPECLIPChannel, RDPEDYCChannel]
		self.vchannels = {
			'ECHO' : VchannelECHO(),
			#'DATATEST1' : VchannelTEST(),
		}
		# Authentication protocols supported
		self.supported_protocols = None # supported_protocols if None: it will be determined automatically. otherwise  select one or more from these SUPP_PROTOCOLS.RDP | SUPP_PROTOCOLS.SSL |SUPP_PROTOCOLS.HYBRID_EX

		# Video settings used by both RDP and VNC
		self.video_width = 1024
		self.video_height = 768
		self.video_out_format = VIDEO_FORMAT.QT5 # the recatngle image format which will be sent via the queue
		
		# Video settings used only by RDP
		# minimum supported BPP
		self.video_bpp_min = 16 
		# maximum supported BPP
		self.video_bpp_max = 16 #max supported bpp
		# all supported BPPs
		self.video_bpp_supported = [15, 16, 24, 32]

		#Performance booster flags
		self.performance_flags = PERF.DISABLE_WALLPAPER | PERF.DISABLE_THEMING | PERF.DISABLE_CURSORSETTINGS | PERF.DISABLE_MENUANIMATIONS | PERF.DISABLE_FULLWINDOWDRAG
		
		# Keyboard settings
		# Keyborad settings used by both RDP and VNC		
		# clinet_keyboard is used by VNC where the keyboard state must be virtualized client-side
		self.client_keyboard = 'enus'

		# Keyboard setting used only by RDP
		# TODO: this should be cleaned up
		self.keyboard_layout = 1033
		self.keyboard_type = 4
		self.keyboard_subtype = 0
		self.keyboard_functionkey = 12

		# Clipboard settings

		# pyperclip is currently the only somewhat working cross-platform clipboard manager
		# disabling it means that clipboard messages will only be dispatched via the external queue
		# but the client's clipboard will not be updated
		self.clipboard_use_pyperclip = True

		# store all incloing clipboard data recieved from server. Text only
		# set it to true to have a file created, or to a string with a file path
		# 
		self.clipboard_store_data = None

		# determines how often the client polls for desktop changes
		self.vnc_fps = 10

		# determines what type of encoding the client supports.
		# Feel free to change the order of these but not the value. 
		# Each value corresponds to an encoding type and only these three anre implemented currently.
		# The order signifies the preference to the server but the server can decide to ignore it
		self.vnc_encodings = [2, 1, 0]
		



import io
from typing import List
import enum
from aardwolf.protocol.T124.userdata.constants import TS_UD_TYPE, MonitorConfig

class TS_MONITOR_DEF:
	def __init__(self):
		self.left:int = None
		self.top:int = None
		self.right:int = None
		self.bottom:int = None
		self.flags:MonitorConfig = None
		
	def to_bytes(self):
		t = self.left.to_bytes(4, byteorder='little', signed = True)
		t += self.top.to_bytes(4, byteorder='little', signed = True)
		t += self.right.to_bytes(4, byteorder='little', signed = True)
		t += self.bottom.to_bytes(4, byteorder='little', signed = True)
		t += self.flags.to_bytes(4, byteorder='little', signed = False)
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_MONITOR_DEF.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_MONITOR_DEF()
		msg.left = int.from_bytes(buff.read(4), byteorder='little', signed = True)
		msg.top = int.from_bytes(buff.read(4), byteorder='little', signed = True)
		msg.right = int.from_bytes(buff.read(4), byteorder='little', signed = True)
		msg.bottom = int.from_bytes(buff.read(4), byteorder='little', signed = True)
		msg.flags = MonitorConfig(int.from_bytes(buff.read(4), byteorder='little', signed = False))
		return msg

	def __repr__(self):
		t = '==== TS_MONITOR_DEF ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-rdpbcgr/6b58e11e-a32b-4903-b736-339f3cfe46ec
class TS_UD_CS_MONITOR:
	def __init__(self):
		self.type:TS_UD_TYPE = TS_UD_TYPE.CS_MONITOR
		self.length:int = None
		self.monitorCount:int = None
		self.monitorDefArray:List[TS_MONITOR_DEF] = []
		
	def to_bytes(self):
		def finish(t):
			t = (len(t)+4).to_bytes(2, byteorder='little', signed = False) + t
			t = self.type.value.to_bytes(2, byteorder='little', signed = False) + t
			return t
		t = len(self.monitorDefArray).value.to_bytes(4, byteorder='little', signed = False)
		for cd in self.monitorDefArray:
			t += cd.to_bytes()
		return finish(t)

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_UD_CS_MONITOR.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_UD_CS_MONITOR()
		msg.type = TS_UD_TYPE(int.from_bytes(buff.read(2), byteorder='little', signed = False))
		msg.length = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.channelCount = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		for _ in range(msg.channelCount):
			msg.monitorDefArray.append(TS_MONITOR_DEF.from_buffer(buff))
		return msg

	def __repr__(self):
		t = '==== TS_UD_CS_NET ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t


import io
import enum

class DRAW_NINEGRID(enum.IntFlag):
	NO_SUPPORT = 0x00000000 #NineGrid bitmap caching and rendering is not supported.
	SUPPORTED = 0x00000001 #Revision 1 NineGrid bitmap caching and rendering is supported. The Revision 1 versions of the stream bitmap alternate secondary orders (see section 2.2.2.2.1.3.5) MUST be used to send the NineGrid bitmap from server to client.
	SUPPORTED_REV2 = 0x00000002 #Revision 2 NineGrid bitmap caching and rendering is supported. The Revision 2 versions of the stream bitmap alternate secondary orders (see section 2.2.2.2.1.3.5) MUST be used to send the NineGrid bitmap from server to client.

class TS_DRAW_NINEGRID_CAPABILITYSET:
	def __init__(self):
		self.drawNineGridSupportLevel:DRAW_NINEGRID = None
		self.drawNineGridCacheSize:int = None
		self.drawNineGridCacheEntries:int = None

	def to_bytes(self):
		t = self.drawNineGridSupportLevel.to_bytes(4, byteorder='little', signed = False)
		t += self.drawNineGridCacheSize.to_bytes(2, byteorder='little', signed = False)
		t += self.drawNineGridCacheEntries.to_bytes(2, byteorder='little', signed = False)
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_DRAW_NINEGRID_CAPABILITYSET.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_DRAW_NINEGRID_CAPABILITYSET()
		msg.drawNineGridSupportLevel = DRAW_NINEGRID(int.from_bytes(buff.read(4), byteorder='little', signed = False))
		msg.drawNineGridCacheSize = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.drawNineGridCacheEntries = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		return msg

	def __repr__(self):
		t = '==== TS_DRAW_NINEGRID_CAPABILITYSET ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t



import enum


class EnumResultStatus(enum.Enum):
	RESULT = 'RESULT'
	FINISHED = 'FINISED'
	ERROR = 'ERROR'

class EnumResult:
	def __init__(self, target_id, target, result, error = None, status = EnumResultStatus.RESULT):
		self.target_id = target_id
		self.target = target
		self.error = error
		self.result = result
		self.status = status

class EnumProgress:
	def __init__(self, total_targets, total_finished, gens_finished, current_finished):
		self.total_targets = total_targets
		self.total_finished = total_finished
		self.gens_finished = gens_finished
		self.current_finished = current_finished


import enum
import io

class SEC_HDR_FLAG(enum.IntFlag):
	EXCHANGE_PKT = 0x0001 #Indicates that the packet is a Security Exchange PDU (section 2.2.1.10). This packet type is sent from client to server only. The client only sends this packet if it will be encrypting further communication and Standard RDP Security mechanisms (section 5.3) are in effect.
	TRANSPORT_REQ = 0x0002 #Indicates that the packet is an Initiate Multitransport Request PDU (section 2.2.15.1). This flag MUST NOT be present if the PDU containing the security header is being sent from client to server. This flag MUST NOT be present if the PDU containing the security header is not being sent on the MCS message channel. The ID of the message channel is specified in the Server Message Channel Data (section 2.2.1.4.5).
	TRANSPORT_RSP = 0x0004 #Indicates that the packet is an Initiate Multitransport Response PDU (section 2.2.15.2). This flag MUST NOT be present if the PDU containing the security header is being sent from server to client. This flag MUST NOT be present if the PDU containing the security header is not being sent on the MCS message channel. The ID of the message channel is specified in the Server Message Channel Data (section 2.2.1.4.5).
	ENCRYPT = 0x0008 #Indicates that the packet is encrypted.
	RESET_SEQNO = 0x0010 #This flag is not processed by any RDP clients or servers and MUST be ignored.
	IGNORE_SEQNO = 0x0020 #This flag is not processed by any RDP clients or servers and MUST be ignored.
	INFO_PKT = 0x0040 #Indicates that the packet is a Client Info PDU (section 2.2.1.11). This packet type is sent from client to server only. If Standard RDP Security mechanisms are in effect, then this packet MUST also be encrypted.
	LICENSE_PKT = 0x0080 #Indicates that the packet is a Licensing PDU (section 2.2.1.12).
	LICENSE_ENCRYPT_CS = 0x0200 #Indicates to the client that the server is capable of processing encrypted licensing packets. It is sent by the server together with any licensing PDUs (section 2.2.1.12).
	LICENSE_ENCRYPT_SC = 0x0200 #Indicates to the server that the client is capable of processing encrypted licensing packets. It is sent by the client together with the SEC_EXCHANGE_PKT flag when sending a Security Exchange PDU (section 2.2.1.10).
	REDIRECTION_PKT = 0x0400 # Indicates that the packet is a Standard Security Server Redirection PDU (section 2.2.13.2.1) and that the PDU is encrypted.
	SECURE_CHECKSUM = 0x0800 #Indicates that the MAC for the PDU was generated using the "salted MAC generation" technique (section 5.3.6.1.1). If this flag is not present, then the standard technique was used (sections 2.2.8.1.1.2.2 and 2.2.8.1.1.2.3).
	AUTODETECT_REQ = 0x1000 #Indicates that the packet is an Auto-Detect Request PDU (section 2.2.14.3). This flag MUST NOT be present if the PDU containing the security header is being sent from client to server This flag MUST NOT be present if the PDU containing the security header is not being sent on the MCS message channel. The ID of the message channel is specified in the Server Message Channel Data (section 2.2.1.4.5).
	AUTODETECT_RSP = 0x2000 # Indicates that the packet is an Auto-Detect Response PDU (section 2.2.14.4). This flag MUST NOT be present if the PDU containing the security header is being sent from server to client This flag MUST NOT be present if the PDU containing the security header is not being sent on the MCS message channel. The ID of the message channel is specified in the Server Message Channel Data (2.2.1.4.5).
	HEARTBEAT = 0x4000 #Indicates that the packet is a Heartbeat PDU (section 2.2.16.1). This flag MUST NOT be present if the PDU containing the security header is not being sent on the MCS message channel. The ID of the message channel is specified in the Server Message Channel Data (2.2.1.4.5).
	FLAGSHI_VALID = 0x8000 #Indicates that the flagsHi field contains valid data. If this flag is not set, then the contents of the flagsHi field MUST be ignored.

class TS_SECURITY_HEADER:
	def __init__(self):
		self.flags:SEC_HDR_FLAG = 0
		self.flagsHi:int = 0

	def to_bytes(self):
		t = self.flags.to_bytes(2, byteorder='little', signed = False)
		t += self.flagsHi.to_bytes(2, byteorder='little', signed = False)
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_SECURITY_HEADER.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_SECURITY_HEADER()
		msg.flags = SEC_HDR_FLAG(int.from_bytes(buff.read(2), byteorder='little', signed = False))
		msg.flagsHi = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		return msg

	def __repr__(self):
		t = '==== TS_SECURITY_HEADER ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t

class TS_SECURITY_HEADER1:
	def __init__(self):
		self.flags:SEC_HDR_FLAG = None
		self.flagsHi:int = 0
		self.dataSignature:bytes = None

	def to_bytes(self):
		t = self.flags.to_bytes(2, byteorder='little', signed = False)
		t += self.flagsHi.to_bytes(2, byteorder='little', signed = False)
		t += self.dataSignature
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_SECURITY_HEADER1.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_SECURITY_HEADER1()
		msg.flags = SEC_HDR_FLAG(int.from_bytes(buff.read(2), byteorder='little', signed = False))
		msg.flagsHi = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.dataSignature = buff.read(8)
		return msg

	def __repr__(self):
		t = '==== TS_SECURITY_HEADER1 ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t

class TS_SECURITY_HEADER2:
	def __init__(self):
		self.flags:SEC_HDR_FLAG = None
		self.flagsHi:int = 0
		self.length:int = 16
		self.version: int = 1
		self.padlen:int = 0
		self.dataSignature:bytes = None

	def to_bytes(self):
		t = self.flags.to_bytes(2, byteorder='little', signed = False)
		t += self.flagsHi.to_bytes(2, byteorder='little', signed = False)
		t += self.length.to_bytes(2, byteorder='little', signed = False)
		t += self.version.to_bytes(1, byteorder='little', signed = False)
		t += self.padlen.to_bytes(1, byteorder='little', signed = False)
		t += self.dataSignature
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_SECURITY_HEADER2.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_SECURITY_HEADER2()
		msg.flags = SEC_HDR_FLAG(int.from_bytes(buff.read(2), byteorder='little', signed = False))
		msg.flagsHi = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.length = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.version = int.from_bytes(buff.read(1), byteorder='little', signed = False)
		msg.padlen = int.from_bytes(buff.read(1), byteorder='little', signed = False)
		msg.dataSignature = buff.read(8)
		return msg

	def __repr__(self):
		t = '==== TS_SECURITY_HEADER2 ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t


import enum

class TS_UD_TYPE(enum.Enum):
	CS_CORE = 0xC001 #The data block that follows contains Client Core Data (section 2.2.1.3.2).
	CS_SECURITY = 0xC002 #The data block that follows contains Client Security Data (section 2.2.1.3.3).
	CS_NET = 0xC003 #The data block that follows contains Client Network Data (section 2.2.1.3.4).
	CS_CLUSTER = 0xC004 #The data block that follows contains Client Cluster Data (section 2.2.1.3.5).
	CS_MONITOR = 0xC005 #The data block that follows contains Client Monitor Data (section 2.2.1.3.6).
	CS_MCS_MSGCHANNEL = 0xC006 #The data block that follows contains Client Message Channel Data (section 2.2.1.3.7).
	CS_MONITOR_EX = 0xC008 #The data block that follows contains Client Monitor Extended Data (section 2.2.1.3.9).
	CS_MULTITRANSPORT = 0xC00A #The data block that follows contains Client Multitransport Channel Data (section 2.2.1.3.8).
	SC_CORE = 0x0C01 #The data block that follows contains Server Core Data (section 2.2.1.4.2).
	SC_SECURITY = 0x0C02 #The data block that follows contains Server Security Data (section 2.2.1.4.3).
	SC_NET = 0x0C03 #The data block that follows contains Server Network Data (section 2.2.1.4.4).
	SC_MCS_MSGCHANNEL = 0x0C04 #The data block that follows contains Server Message Channel Data (section 2.2.1.4.5).
	SC_MULTITRANSPORT = 0x0C08 #The data block that follows contains Server Multitransport Channel Data (section 2.2.1.4.6).

class COLOR_DEPTH(enum.Enum):
	COLOR_4BPP = 0xCA00 #4 bits-per-pixel (bpp)
	COLOR_8BPP = 0xCA01 #8 bpp
	COLOR_16BPP_555 = 0xCA02 #15-bit 555 RGB mask (5 bits for red, 5 bits for green, and 5 bits for blue)
	COLOR_16BPP_565 = 0xCA03 #16-bit 565 RGB mask (5 bits for red, 6 bits for green, and 5 bits for blue)
	COLOR_24BPP = 0xCA04 #24-bit RGB mask (8 bits for red, 8 bits for green, and 8 bits for blue)

class HIGH_COLOR_DEPTH(enum.Enum):
	HIGH_COLOR_4BPP= 0x0004 #4 bpp
	HIGH_COLOR_8BPP = 0x0008 #8 bpp
	HIGH_COLOR_15BPP = 0x000F #15-bit 555 RGB mask (5 bits for red, 5 bits for green, and 5 bits for blue)
	HIGH_COLOR_16BPP = 0x0010 #16-bit 565 RGB mask (5 bits for red, 6 bits for green, and 5 bits for blue)
	HIGH_COLOR_24BPP = 0x0018 #24-bit RGB mask (8 bits for red, 8 bits for green, and 8 bits for blue)

class SUPPORTED_COLOR_DEPTH(enum.IntFlag):
	RNS_UD_24BPP_SUPPORT = 0x0001 #24-bit RGB mask (8 bits for red, 8 bits for green, and 8 bits for blue)
	RNS_UD_16BPP_SUPPORT = 0x0002 #16-bit 565 RGB mask (5 bits for red, 6 bits for green, and 5 bits for blue)
	RNS_UD_15BPP_SUPPORT = 0x0004 #15-bit 555 RGB mask (5 bits for red, 5 bits for green, and 5 bits for blue)
	RNS_UD_32BPP_SUPPORT = 0x0008 #32-bit RGB mask (8 bits for the alpha channel, 8 bits for red, 8 bits for green, and 8 bits for blue)

class RNS_UD_CS(enum.IntFlag):
	SUPPORT_ERRINFO_PDU = 0x0001 #Indicates that the client supports the Set Error Info PDU (section 2.2.5.1).
	WANT_32BPP_SESSION = 0x0002 #Indicates that the client is requesting a session color depth of 32 bpp. This flag is necessary because the highColorDepth field does not support a value of 32. If this flag is set, the highColorDepth field SHOULD be set to 24 to provide an acceptable fallback for the scenario where the server does not support 32 bpp color.
	SUPPORT_STATUSINFO_PDU = 0x0004 #Indicates that the client supports the Server Status Info PDU (section 2.2.5.2).
	STRONG_ASYMMETRIC_KEYS = 0x0008 #Indicates that the client supports asymmetric keys larger than 512 bits for use with the Server Certificate (section 2.2.1.4.3.1) sent in the Server Security Data block (section 2.2.1.4.3).
	UNUSED = 0x0010 #An unused flag that MUST be ignored by the server.
	VALID_CONNECTION_TYPE = 0x0020 #Indicates that the connectionType field contains valid data.
	SUPPORT_MONITOR_LAYOUT_PDU = 0x0040 #Indicates that the client supports the Monitor Layout PDU (section 2.2.12.1).
	SUPPORT_NETCHAR_AUTODETECT = 0x0080 #Indicates that the client supports network characteristics detection using the structures and PDUs described in section 2.2.14.
	SUPPORT_DYNVC_GFX_PROTOCOL = 0x0100 #Indicates that the client supports the Graphics Pipeline Extension Protocol described in [MS-RDPEGFX] sections 1, 2, and 3.
	SUPPORT_DYNAMIC_TIME_ZONE = 0x0200 #Indicates that the client supports Dynamic DST. Dynamic DST information is provided by the client in the cbDynamicDSTTimeZoneKeyName, dynamicDSTTimeZoneKeyName and dynamicDaylightTimeDisabled fields of the Extended Info Packet (section 2.2.1.11.1.1.1).
	SUPPORT_HEARTBEAT_PDU = 0x0400 #Indicates that the client supports the Heartbeat PDU (section 2.2.16.1)

class CONNECTION_TYPE(enum.Enum):
	UNK = 0x00 #custom
	MODEM = 0x01 #Modem (56 Kbps)
	BROADBAND_LOW = 0x02 #Low-speed broadband (256 Kbps - 2 Mbps)
	SATELLITE = 0x03 #Satellite (2 Mbps - 16 Mbps with high latency)
	BROADBAND_HIGH = 0x04 #High-speed broadband (2 Mbps - 10 Mbps)
	WAN = 0x05 #WAN (10 Mbps or higher with high latency)
	LAN = 0x06 #LAN (10 Mbps or higher)
	AUTODETECT = 0x07 #The server SHOULD attempt to detect the connection type. If the connection type can be successfully determined then the performance flags, sent by the client in the performanceFlags field of the Extended Info Packet (section 2.2.1.11.1.1.1), SHOULD be ignored and the server SHOULD determine the appropriate set of performance flags to apply to the remote session (based on the detected connection type). If the RNS_UD_CS_SUPPORT_NETCHAR_AUTODETECT (0x0080) flag is not set in the earlyCapabilityFlags field, then this value SHOULD be ignored.

class ORIENTATION(enum.Enum):
	LANDSCAPE = 0 #The desktop is not rotated.
	PORTRAIT = 90 #The desktop is rotated clockwise by 90 degrees.
	LANDSCAPE_FLIPPED = 180 #The desktop is rotated clockwise by 180 degrees.
	PORTRAIT_FLIPPED = 270 #The desktop is rotated clockwise by 270 degrees.

class ENCRYPTION_FLAG(enum.Enum):
	FRENCH = 0x00000000
	BIT_40 = 0x00000001 #40-bit session keys MUST be used to encrypt data (with RC4) and generate Message Authentication Codes (MAC).
	BIT_128 = 0x00000002 #128-bit session keys MUST be used to encrypt data (with RC4) and generate MACs.
	BIT_56 = 0x00000008 #56-bit session keys MUST be used to encrypt data (with RC4) and generate MACs.
	FIPS = 0x00000010 #All encryption and Message Authentication Code generation routines MUST be Federal Information Processing Standard (FIPS) 140-1 compliant.

class ENCRYPTION_LEVEL(enum.Enum):
	NONE = 0x00000000
	LOW = 0x00000001
	CLIENT_COMPATIBLE = 0x00000002
	HIGH = 0x00000003
	FIPS = 0x00000004

class ChannelOption(enum.IntFlag):
	INITIALIZED =0x80000000 #This flag is unused and its value MUST be ignored by the server.
	ENCRYPT_RDP = 0x40000000 #This flag is unused and its value MUST be ignored by the server.
	ENCRYPT_SC = 0x20000000 #This flag is unused and its value MUST be ignored by the server.
	ENCRYPT_CS = 0x10000000 #This flag is unused and its value MUST be ignored by the server.
	PRI_HIGH = 0x08000000 #Channel data MUST be sent with high MCS priority.
	PRI_MED = 0x04000000 #Channel data MUST be sent with medium MCS priority.
	PRI_LOW = 0x02000000 #Channel data MUST be sent with low MCS priority.
	COMPRESS_RDP = 0x00800000 #Virtual channel data MUST be compressed if RDP data is being compressed.
	COMPRESS = 0x00400000 #Virtual channel data MUST be compressed, regardless of RDP compression settings.
	SHOW_PROTOCOL = 0x00200000 #The value of this flag MUST be ignored by the server. The visibility of the Channel PDU Header (section 2.2.6.1.1) is determined by the CHANNEL_FLAG_SHOW_PROTOCOL (0x00000010) flag as defined in the flags field (section 2.2.6.1.1).
	REMOTE_CONTROL_PERSISTENT = 0x00100000 #Channel MUST be persistent across remote control transactions.

class ClusterInfo(enum.IntFlag):
	REDIRECTION_SUPPORTED = 0x00000001 #The client can receive server session redirection packets. If this flag is set, the ServerSessionRedirectionVersionMask MUST contain the server session redirection version that the client supports.
	ServerSessionRedirectionVersionMask = 0x0000003C #The server session redirection version that the client supports. See the discussion which follows this table for more information.
	REDIRECTED_SESSIONID_FIELD_VALID = 0x00000002 #The RedirectedSessionID field contains an ID that identifies a session on the server to associate with the connection.
	REDIRECTED_SMARTCARD = 0x00000040 #The client logged on with a smart card.

class MonitorConfig(enum.IntFlag):
	TS_MONITOR_PRIMARY = 0x00000001

class MultitransportFlags(enum.IntFlag):
	TRANSPORTTYPE_UDPFECR = 0x01 #RDP-UDP Forward Error Correction (FEC) reliable transport ([MS-RDPEUDP] sections 1 to 3).
	TRANSPORTTYPE_UDPFECL = 0x04 #RDP-UDP FEC lossy transport ([MS-RDPEUDP] sections 1 to 3).
	TRANSPORTTYPE_UDP_PREFERRED = 0x100 #Indicates that tunneling of static virtual channel traffic over UDP is supported ([MS-RDPEDYC] section 3.1.5.4).
	SOFTSYNC_TCP_TO_UDP = 0x200 #Indicates that switching dynamic virtual channels from the TCP to UDP transport is supported ([MS-RDPEDYC] section 3.1.5.3).

class RNS_UD_SC(enum.IntFlag):
	EDGE_ACTIONS_SUPPORTED_V1 = 0x00000001
	DYNAMIC_DST_SUPPORTED = 0x00000002
	EDGE_ACTIONS_SUPPORTED_V2 = 0x00000004


import os
import json
import base64
import traceback
import xml.etree.ElementTree as ET

from tqdm import tqdm

from aardwolf.keyboard import KeyboardLayout, VK_MODIFIERS

"""
This script will fetch all keyboard layouts from https://kbdlayout.info/ and parse the results.
Results will be stored in a JSON format to be consumed by the 'KeyboardLayout' class.
As I do not control the website, I can't guarantee that this script will work in the future.

"""

def full_refresh():
    loc_info = {}
    import re
    import urllib.request, urllib.error, urllib.parse
    
    klids_to_filename = {}
    names_to_filename = {}
    shortnames_to_filename = {}


    dirname = os.path.dirname(__file__)
    layout_dir_path = os.path.join(dirname,'layouts/')
    layout_manager_py_path = os.path.join(dirname,'layoutmanager.py')
    with open(layout_dir_path + '__init__.py' , 'w') as f:
        f.write('')
    


    dlocations = {}
    url = 'https://kbdlayout.info/'
    response = urllib.request.urlopen(url)
    webContent = response.read().decode('UTF-8')
    
    #print(webContent)
    hrefs = re.findall(r'href=[\'"]?([^\'" >]+)', webContent)
    for href in hrefs:
        if href.startswith('/00') is True or href.startswith('/k') is True or href.startswith('/K') is True:
            dlocations[href[1:-1]] = 1

    pbar = tqdm(dlocations)
    i = 0
    for loc in pbar:
        #i += 1
        #if i == 10:
        #    break
        try:
            iurl = url + '/' + loc +'/'
            #print('Fetching url: %s' % kurl)
            response = urllib.request.urlopen(iurl)
            webContent = response.read().decode('UTF-8')
        
            match = re.findall(r'(?<=<th>KLID</th><td>)(.*)(?=</td>|$)', webContent)
            if match != '':
                pbar.set_description('Processing %s' % ','.join(match)[:50].ljust(50))
                loc_info[loc] = match        
            else:
                loc_info[loc] = []

            murl = url + '/' + loc +'/download/cldr'
            response = urllib.request.urlopen(murl)
            webContent = response.read().decode('UTF-8')
            root = ET.fromstring(webContent)
            locale = ''
            names = []
            if 'locale' in root.attrib:
                locale = root.attrib['locale']
            for entry in root:
                if entry.tag == 'names':
                    for name in entry:
                        if name.tag == 'name':
                            if 'value' in name.attrib:
                                names.append(name.attrib['value'])

            kurl = url + '/' + loc +'/download/xml'
            #print('Fetching url: %s' % kurl)
            response = urllib.request.urlopen(kurl)
            webContent = response.read().decode('UTF-8')
            #print('Parsing XML data...')
            layout = kbdlayout_xml_to_kayboardlayout(loc, loc_info[loc], locale, names, webContent)
            fname = 'layout_%s.py' % loc
            layouts_file_path = os.path.join(layout_dir_path,fname)
            
            with open(layouts_file_path, 'w', newline = '') as f:
                layoutdata = 'layoutdata = "%s"\r\n' % base64.b64encode(layout.to_json().encode()).decode()
                f.write(layoutdata)
            

            for klid in loc_info[loc]:
                shortname = klid.split(' ',1)[1][1:-1].replace('-','').lower()
                klids_to_filename[klid] = fname
                shortnames_to_filename[shortname] = fname

            for name in names:
                names_to_filename[name] = fname
            

        except Exception as e:
            print('Error! Parsing failed: %s' % loc)
            traceback.print_exc()
            continue
        
    
    with open(layout_manager_py_path, 'w') as f:
        f.write("""
import json
import base64
import importlib

from aardwolf.keyboard import KeyboardLayout

### Importlib is used here, as there are many keyboard layouts and you probably don't wish to use all of them
### Importing all of them would take a considerable time however, hence dynamic loading

class KeyboardLayoutManager:
    def __init__(self):
        self.klids_to_filename = json.loads(base64.b64decode("%s").decode())
        self.names_to_filename = json.loads(base64.b64decode("%s").decode())
        self.shortnames_to_filename = json.loads(base64.b64decode("%s").decode())
    
    def __layout_loader(self, filename):
        layoutdata = importlib.import_module('aardwolf.keyboard.layouts.%%s' %% filename[:-3])
        return KeyboardLayout.from_layoutdata(layoutdata.layoutdata)


    def get_layout_by_klid(self, klid):
        if klid not in self.klids_to_filename:
            return None
        return self.__layout_loader(self.klids_to_filename[klid])

    def get_layout_by_name(self, name):
        if name not in self.names_to_filename:
            return None
        
        return self.__layout_loader(self.names_to_filename[name]) 
    
    def get_layout_by_shortname(self, name):
        if name not in self.shortnames_to_filename:
            return None
        return self.__layout_loader(self.shortnames_to_filename[name]) 
    
    def get_klids(self):
        for klid in self.klids_to_filename:
            yield klid
    
    def get_names(self):
        for name in self.names_to_filename:
            yield name
    
    def get_shortnames(self):
        for name in self.shortnames_to_filename:
            yield name

if __name__ == '__main__':
    kl = KeyboardLayoutManager()
    kl.get_layout_by_shorname('sq')
""" % (  base64.b64encode(json.dumps(klids_to_filename).encode()).decode(), base64.b64encode(json.dumps(names_to_filename).encode()).decode(), base64.b64encode(json.dumps(shortnames_to_filename).encode()).decode() ))



def kbdlayout_xml_to_kayboardlayout(loc, klids, locale, layout_names, xml_data):
    layout = KeyboardLayout()
    layout.klid = klids
    layout.name = layout_names
    root = ET.fromstring(xml_data)
    #root = tree.getroot()
    #print(root)
    #print(root.tag)
    #print(root.attrib)
    for k in root.attrib:
        if k == 'RightAltIsAltGr':
            layout.RightAltIsAltGr = True if root.attrib[k].lower() == 'true' else False
        elif k == 'ShiftCancelsCapsLock':
            layout.ShiftCancelsCapsLock = True if root.attrib[k].lower() == 'true' else False
        elif k == 'ChangesDirectionality':
            layout.ChangesDirectionality = True if root.attrib[k].lower() == 'true' else False
        else:
            print('Not porcessed attrib key: %s' % k)
            input()
    for child in root:
        for entry in child:
            if entry.tag == 'PK':
                sc = int.from_bytes(bytes.fromhex((entry.attrib['SC'])), byteorder='big', signed = False)
                layout.sc_to_vk[sc] = entry.attrib['VK']
                
                for result in entry:
                    if result.tag == 'Result':
                        extra_with = []
                        if 'With' in result.attrib:
                            extra_with = result.attrib['With'].split(' ')
                        children = list(result.getchildren())
                        for rc in children:
                            if rc.tag == 'DeadKeyTable':
                                process_tag(layout, sc, rc, extra_with)
                        
                        process_tag(layout, sc, result, [])
                        #if 'Text' in result.attrib:
                        #    data = result.attrib['Text']
                        #if 'TextCodepoints' in result.attrib:
                        #    data = result.attrib['TextCodepoints']
                        #
                        #modifiers = 0
                        #if 'With' in result.attrib:
                        #    for wtag in result.attrib['With'].split(' '):
                        #        modifiers |= VK_MODIFIERS[wtag]
                        #
                        #if sc not in layout.sc_to_char:
                        #    layout.sc_to_char[sc] = {}
                        #
                        #if data is not None:
                        #    layout.sc_to_char[sc][modifiers] = data
                        #    layout.char_to_sc[data] = (sc, '')
                        #    if modifiers != 0:
                        #        layout.char_to_sc[data] = (sc, modifiers)
                            
                        
            else:
                print('Unknown tag %s' % entry.tag)
                input()

    return layout

def process_tag(layout, sc, result, extra_with):
    data = None
    if 'Text' in result.attrib:
        data = result.attrib['Text']
    if 'Accent' in result.attrib:
        data = result.attrib['Accent']
    if 'TextCodepoints' in result.attrib:
        data = result.attrib['TextCodepoints']
    
    modifiers = 0
    for wtag in extra_with:
        modifiers |= VK_MODIFIERS[wtag]
    if 'With' in result.attrib:
        for wtag in result.attrib['With'].split(' '):
            modifiers |= VK_MODIFIERS[wtag]
    
    if sc not in layout.sc_to_char:
        layout.sc_to_char[sc] = {}
    
    if data is not None:
        layout.sc_to_char[sc][modifiers] = data
        layout.char_to_sc[data] = (sc, '')
        if modifiers != 0:
            layout.char_to_sc[data] = (sc, modifiers)

if __name__ == '__main__':
    full_refresh()
    #pass


import io
import enum
from typing import List
from aardwolf.protocol.T128.share import TS_SHAREDATAHEADER
from aardwolf.protocol.pdu.input import TS_INPUT_EVENT


class TS_INPUT_PDU_DATA:
	def __init__(self):
		self.shareDataHeader:TS_SHAREDATAHEADER = None
		self.numEvents:int = 0
		self.pad2Octets:int = 0
		self.slowPathInputEvents:List[TS_INPUT_EVENT] = []

	def to_bytes(self):
		#t  = self.shareDataHeader.to_bytes()
		data = b''
		for inputevt in self.slowPathInputEvents:
			data += inputevt.to_bytes()
		t = len(self.slowPathInputEvents).to_bytes(2, byteorder='little', signed = False)
		t += self.pad2Octets.to_bytes(2, byteorder='little', signed = False)
		t += data
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_INPUT_PDU_DATA.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_INPUT_PDU_DATA()
		msg.shareDataHeader = TS_SHAREDATAHEADER.from_buffer(buff)
		msg.numEvents = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.pad2Octets = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		for _ in range(msg.numEvents):
			msg.slowPathInputEvents.append(TS_INPUT_EVENT.from_buffer(buff))
		return msg

	def __repr__(self):
		t = '==== TS_INPUT_PDU_DATA ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t


import io
import enum
from aardwolf.protocol.T128.share import TS_SHAREDATAHEADER

class TS_SYNCHRONIZE_PDU:
	def __init__(self):
		self.shareDataHeader:TS_SHAREDATAHEADER = None
		self.messageType:int = 1
		self.targetUser:int = None #MCS channel ID

	def to_bytes(self):
		t = self.messageType.to_bytes(2, byteorder='little', signed = False)
		t += self.targetUser.to_bytes(2, byteorder='little', signed = False)
		#self.shareDataHeader.shareControlHeader.totalLength = 6 + len(t)
		#t = self.shareControlHeader.to_bytes() + t
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_SYNCHRONIZE_PDU.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_SYNCHRONIZE_PDU()
		msg.shareDataHeader = TS_SHAREDATAHEADER.from_buffer(buff)
		msg.messageType = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.targetUser = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		return msg

	def __repr__(self):
		t = '==== TS_SYNCHRONIZE_PDU ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t


import io
from aardwolf.extensions.RDPEDYC.protocol import DYNVC_CMD, dynvc_header_to_bytes, dynvc_header_from_buff

# https://learn.microsoft.com/en-us/openspecs/windows_protocols/ms-rdpedyc/4448ba4d-9a72-429f-8b65-6f4ec44f2985
class DYNVC_CREATE_REQ:
    def __init__(self):
        self.cbid:int = None
        self.pri:int   = 0
        self.cmd:DYNVC_CMD = DYNVC_CMD.CREATE_RSP
        self.ChannelId:int   = 0
        self.ChannelName:str = None

    @staticmethod
    def from_bytes(data: bytes):
        return DYNVC_CREATE_REQ.from_buffer(io.BytesIO(data))

    @staticmethod
    def from_buffer(buff: io.BytesIO):
        msg = DYNVC_CREATE_REQ()
        msg.cbid, msg.pri, msg.cmd = dynvc_header_from_buff(buff, cbid_mod=True)
        msg.ChannelId = int.from_bytes(buff.read(msg.cbid), byteorder='little', signed=False)
        msg.ChannelName = ''
        for _ in range(255):
            t = buff.read(1)
            if t == b'\x00':
                break
            msg.ChannelName += chr(ord(t))
        return msg

    def to_bytes(self):
        if self.cbid is None:
            self.cbid = 2
        if self.ChannelName[:-1] != '\x00':
            self.ChannelName += '\x00'
        
        t = b''
        t += dynvc_header_to_bytes(self.cbid, self.pri, self.cmd, cbid_mod=True)
        t += self.ChannelId.to_bytes(self.cbid, byteorder='little', signed=False)
        t += self.ChannelName.encode()
        return t
    
    def __str__(self):
        t = ''
        for k in self.__dict__:
            t += '%s: %s\r\n' % (k, self.__dict__[k])
        return t


# https://learn.microsoft.com/en-us/openspecs/windows_protocols/ms-rdpedyc/8f284ea3-54f3-4c24-8168-8a001c63b581
class DYNVC_CREATE_RSP:
    def __init__(self):
        self.cbid:int = None
        self.sp:int   = 0
        self.cmd:DYNVC_CMD = DYNVC_CMD.CREATE_RSP
        self.ChannelId:int   = 0
        self.CreationStatus:int = None # actually a HRESULT

    @staticmethod
    def from_bytes(data: bytes):
        return DYNVC_CREATE_RSP.from_buffer(io.BytesIO(data))

    @staticmethod
    def from_buffer(buff: io.BytesIO):
        msg = DYNVC_CREATE_RSP()
        msg.cbid, msg.sp, msg.cmd = dynvc_header_from_buff(buff, cbid_mod=True)
        msg.ChannelId = int.from_bytes(buff.read(msg.cbid), byteorder='little', signed=False)
        msg.CreationStatus = int.from_bytes(buff.read(4), byteorder='little', signed=False)
        return msg

    def to_bytes(self):
        if self.cbid is None:
            self.cbid = 1
        
        t = b''
        t += dynvc_header_to_bytes(self.cbid, self.sp, self.cmd, cbid_mod=True)
        t += self.ChannelId.to_bytes(self.cbid, byteorder='little', signed=False)
        t += self.CreationStatus.to_bytes(4, byteorder='little', signed=False)
        return t

    def __str__(self):
        t = ''
        for k in self.__dict__:
            t += '%s: %s\r\n' % (k, self.__dict__[k])
        return t


import io
import asyncio
from aardwolf.extensions.RDPEDYC.vchannels import VirtualChannelBase

# SOCKS over RDP vchannel
#
# This vchannel is compatible with the "SocksOverRDP" project written by Balazs Bucsay (@xoreipeip)
# 
# To make use of this, you'll need to log in to the server and execute the "SocksOverRDP-Server" component 
# from the following repo https://github.com/nccgroup/SocksOverRDP
# The server component doesn't require administrative privileges, however 
# it will only run until your session is alive. You'll have to re-launch it every time you log in.
#
# Thank you for the cool project Balazs!

class SOCKSDataMessage:
    def __init__(self, connid:int, iserror:bool, data:bytes):
        self.connid = connid
        self.datalen = len(data)
        self.iserror = iserror
        self.data = data

    def to_bytes(self):
        t = self.connid.to_bytes(4, byteorder='little', signed=False)
        t += self.datalen.to_bytes(4, byteorder='little', signed=False)
        t += bytes([int(self.iserror)])
        t += self.data
        return t

    @staticmethod
    def from_bytes(data:bytes):
        return SOCKSDataMessage.from_buffer(io.BytesIO(data))
    
    @staticmethod
    def from_buffer(buff:io.BytesIO):
        connid = int.from_bytes(buff.read(4), byteorder='little', signed=False)
        datalen = int.from_bytes(buff.read(4), byteorder='little', signed=False)
        iserror = bool(buff.read(1)[0])
        data = buff.read(datalen)
        return SOCKSDataMessage(connid, iserror, data)
    
    def __str__(self):
        t = ''
        for k in self.__dict__:
            t += '%s: %s\r\n' % (k, self.__dict__[k])
        return k

class SocksOverRDPChannel(VirtualChannelBase):
    def __init__(self, channelname, listen_ip:str, listen_port:int):
        VirtualChannelBase.__init__(self, channelname)
        self.listen_ip = listen_ip
        self.listen_port = listen_port
        self.__server = None
        self.__connection_id = 1
        self.__connections = {}

    def get_connection_id(self):
        t = self.__connection_id
        self.__connection_id += 1
        return t

    async def __handle_tcp_client(self, reader:asyncio.StreamReader, writer:asyncio.StreamWriter):
        try:
            #print('Client connected!')
            connid = self.get_connection_id()
            self.__connections[connid] = writer
            while True:
                data = await reader.read(1590)
                if data == b'':
                    #print('Client disconnected!')
                    return
                msg = SOCKSDataMessage(connid, False, data)
                await self.channel_data_out(msg.to_bytes())

        except Exception as e:
            print('Error! %s' % e)
            return
        finally:
            msg = SOCKSDataMessage(connid, True, b'AAAA')
            await self.channel_data_out(msg.to_bytes())

    async def channel_init(self):
        #print('Channel init called!')
        self.__server = await asyncio.start_server(self.__handle_tcp_client, self.listen_ip, self.listen_port)
        print('SOCKS server listening on %s:%s' % (self.listen_ip, self.listen_port))
        return True, None

    async def channel_data_in(self, data:bytes):
        try:
            msg = SOCKSDataMessage.from_bytes(data)
            if msg.connid not in self.__connections:
                print('MSG recieved for unknown connid "%s"' % msg.connid)
                return
            if msg.iserror is False:
                self.__connections[msg.connid].write(msg.data)
                await self.__connections[msg.connid].drain()

            else:
                print('Remote end terminated the connection!')
                self.__connections[msg.connid].close()
                del self.__connections[msg.connid]
        except Exception as e:
            print('Error! %s' % e)
            return

    async def channel_closed(self):
        for connid in self.__connections:
            self.__connections[connid].close()
        self.__server.close()




import io
import enum

class FILECONTENTS_FLAG(enum.Enum):
	FILECONTENTS_SIZE = 0x00000001 #A request for the size of the file identified by the lindex field. The size MUST be returned as a 64-bit, unsigned integer. The cbRequested field MUST be set to 0x00000008 and both the nPositionLow and nPositionHigh fields MUST be set to 0x00000000.
	FILECONTENTS_RANGE = 0x00000002 #A request for the data present in the file identified by the lindex field. The data to be retrieved is extracted starting from the offset given by the nPositionLow and nPositionHigh fields. The maximum number of bytes to extract is specified by the cbRequested field.

class CLIPRDR_FILECONTENTS_REQUEST:
	def __init__(self):
		self.streamId:int = None
		self.lindex:int = None
		self.dwFlags:FILECONTENTS_FLAG = None
		self.nPositionLow:int = None
		self.nPositionHigh:int = None
		self.cbRequested:int = None
		self.clipDataId:int = None

		self.nPosition:int = None


	def to_bytes(self):
		if self.nPosition is not None:
			self.nPositionHigh = self.nPosition >> 32
			self.nPositionLow = self.nPosition & 0xFFFFFFFF

		t = self.streamId.to_bytes(4, byteorder='little', signed = False)
		t += self.lindex.to_bytes(4, byteorder='little', signed = False)
		t += self.dwFlags.value.to_bytes(4, byteorder='little', signed = False)
		t += self.nPositionLow.to_bytes(4, byteorder='little', signed = False)
		t += self.nPositionHigh.to_bytes(4, byteorder='little', signed = False)
		t += self.cbRequested.to_bytes(4, byteorder='little', signed = False)
		if self.clipDataId is not None:
			t += self.clipDataId.to_bytes(4, byteorder='little', signed = False)

		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return CLIPRDR_FILECONTENTS_REQUEST.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = CLIPRDR_FILECONTENTS_REQUEST()
		msg.streamId = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.lindex = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.dwFlags = FILECONTENTS_FLAG(int.from_bytes(buff.read(4), byteorder='little', signed = False))
		msg.nPositionLow = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.nPositionHigh = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.cbRequested = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		if buff.read(1) != b'':
			buff.seek(-1,1)
			msg.clipDataId = int.from_bytes(buff.read(4), byteorder='little', signed = False)

		msg.nPosition = msg.nPositionHigh << 32 | msg.nPositionLow
		return msg

	def __repr__(self):
		t = '==== CLIPRDR_FILECONTENTS_REQUEST ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t



import io
import enum

class VCCAPS(enum.IntFlag):
	NO_COMPR = 0x00000000 #Virtual channel compression is not supported.
	COMPR_SC = 0x00000001 #Indicates to the server that virtual channel compression is supported by the client for server-to-client traffic. The highest compression level supported by the client is advertised in the Client Info PDU (section 2.2.1.11).
	COMPR_CS_8K = 0x00000002 #Indicates to the client that virtual channel compression is supported by the server for client-to-server traffic (the compression level is limited to RDP 4.0 bulk compression).

class TS_VIRTUALCHANNEL_CAPABILITYSET:
	def __init__(self):
		self.flags:VCCAPS = VCCAPS.NO_COMPR
		self.VCChunkSize:int = None

	def to_bytes(self):
		t = self.flags.to_bytes(4, byteorder='little', signed = False)
		if self.VCChunkSize is not None:
			t += self.VCChunkSize.to_bytes(4, byteorder='little', signed = False)
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_VIRTUALCHANNEL_CAPABILITYSET.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_VIRTUALCHANNEL_CAPABILITYSET()
		msg.flags = VCCAPS(int.from_bytes(buff.read(4), byteorder='little', signed = False))
		if buff.read(1) == b'':
			return msg
		buff.seek(-1,1)
		#optional
		msg.VCChunkSize = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		return msg

	def __repr__(self):
		t = '==== TS_VIRTUALCHANNEL_CAPABILITYSET ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t


import enum

class RDPDATATYPE(enum.Enum):
	VIDEO = "VIDEO"
	KEYSCAN = 'KEYSCAN'
	KEYUNICODE = 'KEYUNICODE'
	MOUSE = 'MOUSE'
	CLIPBOARD_READY = 'CLIPBOARD_READY'
	CLIPBOARD_DATA_TXT = 'CLIPBOARD_DATA_TXT'
	CLIPBOARD_CONSUMED = 'CLIPBOARD_CONSUMED'
	CLIPBOARD_NEW_DATA_AVAILABLE = 'CLIPBOARD_NEW_DATA_AVAILABLE'
	BEEP = 'BEEP'

from aardwolf.commons.queuedata.video import RDP_VIDEO
from aardwolf.commons.queuedata.keyboard import RDP_KEYBOARD_SCANCODE, RDP_KEYBOARD_UNICODE
from aardwolf.commons.queuedata.mouse import RDP_MOUSE
from aardwolf.commons.queuedata.clipboard import RDP_CLIPBOARD_READY, RDP_CLIPBOARD_DATA_TXT, RDP_CLIPBOARD_CONSUMED, RDP_CLIPBOARD_NEW_DATA_AVAILABLE
from aardwolf.commons.queuedata.beep import RDP_BEEP




import io
import enum
from typing import List

class TS_PALETTE_ENTRY:
	def __init__(self):
		self.red:int = None
		self.green:int = None
		self.blue:int = None

	def to_bytes(self):
		t = self.red.to_bytes(1, byteorder='little', signed=False)
		t += self.green.to_bytes(1, byteorder='little', signed=False)
		t += self.blue.to_bytes(1, byteorder='little', signed=False)
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_PALETTE_ENTRY.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_PALETTE_ENTRY()
		msg.red = int.from_bytes(buff.read(1), byteorder='little', signed=False)
		msg.green = int.from_bytes(buff.read(1), byteorder='little', signed=False)
		msg.blue = int.from_bytes(buff.read(1), byteorder='little', signed=False)
		return msg

	def __repr__(self):
		t = '==== TS_PALETTE_ENTRY ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t

class TS_UPDATE_PALETTE_DATA:
	def __init__(self):
		self.updateType:int = 0x0002
		self.pad2Octets:bytes = b'\x00\x00'
		self.numberColors:int = None
		self.paletteEntries:List[TS_PALETTE_ENTRY] = []

	def to_bytes(self):
		t = self.updateType.to_bytes(2, byteorder='little', signed=False)
		t += self.pad2Octets
		t += len(self.paletteEntries).to_bytes(2, byteorder='little', signed=False)
		for pal in self.paletteEntries:
			t += pal.to_bytes()
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_UPDATE_PALETTE_DATA.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_UPDATE_PALETTE_DATA()
		msg.updateType = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		msg.pad2Octets = buff.read(2)
		msg.numberColors = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		for _ in range(msg.numberColors):
			msg.paletteEntries.append(TS_PALETTE_ENTRY.from_buffer(buff))
		return msg

	def __repr__(self):
		t = '==== TS_UPDATE_PALETTE_DATA ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t





import enum
import json
import base64

class VK_MODIFIERS(enum.IntFlag):
    VK_SHIFT = 1
    VK_CONTROL = 2
    VK_MENU = 4
    VK_CAPITAL = 8
    VK_OEM_8 = 16
    VK_KANA = 32
    VK_NUMLOCK = 64
    VK_WIN = 128

class KeyboardLayout:
    def __init__(self):
        self.klid = None
        self.name = None
        self.RightAltIsAltGr = None 
        self.ShiftCancelsCapsLock = None 
        self.ChangesDirectionality = None
        self.char_to_sc_nomod = {} #char -> sc
        self.char_to_sc = {} #char -> (sc,modifiers)
        self.sc_to_char = {} #scancode -> modifiers -> char
        self.vk_to_sc = {}
        self.sc_to_vk = {}

    def char_to_scancode(self, char):
        if char in self.char_to_sc_nomod:
            return self.char_to_sc_nomod[char], VK_MODIFIERS(0)
        
        return self.char_to_sc[char]

    def scancode_to_char(self, sc, modifiers:VK_MODIFIERS = VK_MODIFIERS(0)):
        """0 is no modifiers present"""
        if sc not in self.sc_to_char:
            return None
        avilable_modifiers = self.sc_to_char[sc]
        if len(avilable_modifiers) == 0:
            return None
        if modifiers in avilable_modifiers:
            return avilable_modifiers[modifiers]
        #print('available modifiers: %s' % avilable_modifiers)
        if modifiers not in avilable_modifiers and VK_MODIFIERS(0) in avilable_modifiers:
            return avilable_modifiers[VK_MODIFIERS(0)]
        return None

    def scancode_to_vk(self, sc):
        if sc not in self.sc_to_vk:
            return None
        return self.sc_to_vk[sc]
    
    def vk_to_scancode(self, vk):
        return self.vk_to_sc[vk]

    def to_dict(self):
        return {
            'klid' : self.klid,
            'name' : self.name,
            'RightAltIsAltGr' : self.RightAltIsAltGr,
            'ShiftCancelsCapsLock' : self.ShiftCancelsCapsLock,
            'ChangesDirectionality' : self.ChangesDirectionality,
            'char_to_sc' : self.char_to_sc,
            'sc_to_char' : self.sc_to_char,
            'vk_to_sc' : self.vk_to_sc,
            'sc_to_vk' : self.sc_to_vk,
        }
    
    def to_json(self):
        return json.dumps(self.to_dict())

    @staticmethod
    def from_dict(d):
        kl = KeyboardLayout()
        kl.klid = d['klid']
        kl.name = d['name']
        kl.RightAltIsAltGr = d['RightAltIsAltGr']
        kl.ShiftCancelsCapsLock = d['ShiftCancelsCapsLock']
        kl.ChangesDirectionality = d['ChangesDirectionality']
        kl.sc_to_char = {}
        kl.char_to_sc_nomod = {}
        for k in d['sc_to_char']:
            sc = int(k)
            kl.sc_to_char[sc] = {}
            for modifier in d['sc_to_char'][k]:
                if modifier == '0':
                    kl.char_to_sc_nomod[d['sc_to_char'][k][modifier]] = sc
                mo = VK_MODIFIERS(int(modifier))
                kl.sc_to_char[sc][mo] = d['sc_to_char'][k][modifier]
        
        kl.char_to_sc = {}
        for k in d['char_to_sc']:
            if k in kl.char_to_sc and kl.char_to_sc[k][1] == '':
                continue

            kl.char_to_sc[k] = d['char_to_sc'][k]
            if kl.char_to_sc[k][1] != '':
                mo = VK_MODIFIERS(int(kl.char_to_sc[k][1]))
                kl.char_to_sc[k] = (kl.char_to_sc[k][0], mo)
            else:
                kl.char_to_sc[k] = (kl.char_to_sc[k][0], mo)

        kl.sc_to_vk = {}
        for k in d['sc_to_vk']:
            kl.sc_to_vk[int(k)] = d['sc_to_vk'][k]

        kl.vk_to_sc = {}
        for k in d['sc_to_vk']:
            if d['sc_to_vk'][k] in kl.vk_to_sc and kl.vk_to_sc[d['sc_to_vk'][k]] < int(k):
                continue
            kl.vk_to_sc[d['sc_to_vk'][k]] = int(k)

        #kl.vk_to_sc = {v: k for k, v in d['sc_to_vk'].items()}

        return kl
    
    @staticmethod
    def from_json(s):
        return KeyboardLayout.from_dict(json.loads(s))
    
    @staticmethod
    def from_layoutdata(layoutdata):
        lj = base64.b64decode(layoutdata).decode()
        return KeyboardLayout.from_json(lj)